#include "../../cuda_VS_header.h"
#include <stdlib.h>
#include <stdio.h>

//Spolu s pepnaem -DDEBUG pi chyb vype i na jakm dku byla tato funkce volna. Vrac true pokud chyba nastala
inline bool check_cuda_errors(const char *filename, const int line_number) {
	cudaThreadSynchronize();
	cudaError_t error = cudaGetLastError();
	if(error != cudaSuccess) {
#ifdef DEBUG
		printf("CUDA error at %s:%i: %s\n", filename, line_number, cudaGetErrorString(error));
#else
		printf("CUDA error %s\n", cudaGetErrorString(error));
#endif
		return true;
	}
	return false;
}

//definice kernelu
__global__ void vector_add(int* a_array, int *b_array, int *result_array) {
	int index = blockIdx.x * blockDim.x + threadIdx.x;
	result_array[index] = a_array[index] + b_array[index];
}

int main(void) {
	int return_number = 0;
	int num_elements = 256;
	int num_bytes = num_elements * sizeof(int);

	//pole v hostovi
	int* a_host_array = 0;
	int* b_host_array = 0;
	int* result_host_array = 0;

	//pole v zazen
	int* a_dev_array = 0;
	int* b_dev_array = 0;
	int* result_dev_array = 0;

	//alokace pamti v hostovi
	a_host_array = (int*)malloc(num_bytes);
	b_host_array = (int*)malloc(num_bytes);
	result_host_array = (int*)malloc(num_bytes);

	//alokace pamti na zazen
	cudaMalloc((void**)&a_dev_array, num_bytes);
	cudaMalloc((void**)&b_dev_array, num_bytes);
	cudaMalloc((void**)&result_dev_array, num_bytes);	
	if(check_cuda_errors(__FILE__, __LINE__)) goto end;

	//nastaven pole a v hostovi
	for(int i=0; i<num_elements; i++) {
		a_host_array[i] = i;
		b_host_array[i] = num_elements-i;
	}

	//pekoprovn a_host_array do zazen
	cudaMemcpy(a_dev_array, a_host_array, num_bytes, cudaMemcpyHostToDevice);
	cudaMemcpy(b_dev_array, b_host_array, num_bytes, cudaMemcpyHostToDevice);
	if(check_cuda_errors(__FILE__, __LINE__)) goto end;


	//zavoln kernelu
	int block_size = 128;
	int grid_size = num_elements / block_size;
	vector_add<<<grid_size,block_size>>>(a_dev_array, b_dev_array, result_dev_array);
	if(check_cuda_errors(__FILE__, __LINE__)) goto end;

	//zskn vsledku
	cudaMemcpy(result_host_array, result_dev_array, num_bytes, cudaMemcpyDeviceToHost);
	if(check_cuda_errors(__FILE__, __LINE__)) goto end;

	//vypsn vsledku
	for(int i=0; i<num_elements; i++) {
		printf("%d ", result_host_array[i]);
	}
	printf("\n");

end:
	//dealokace pol hosta
	free(a_host_array);
	free(b_host_array);
	free(result_host_array);

	//dealokace pol zazen
	cudaFree(a_dev_array);
	cudaFree(b_dev_array);
	cudaFree(result_dev_array);

	return 0;
}